from delune import indexinfo
from rs4 import confparse, logger
from . import classifier
from delune.searcher import searcher
from delune.searcher import memory
import os

class Classifier:
	def __init__ (self, path, debug = False):	
		memory.create (1, 32768, 100, "segment")
		lg = logger.screen_logger ()
		conf = confparse.ConfParse (path)
		
		base_index = os.path.join (os.path.split (conf.getfilename ()) [0], os.path.split (conf.getopt ("common", "base_index"))[-1])		
		si = indexinfo.IndexInfo (confparse.ConfParse (base_index))
		base_schr = searcher.Searcher (si, lg)
		
		inf = indexinfo.IndexInfo (conf)
		self.f = classifier.Classifier (inf, base_schr, lg)
		self.f.initialize ()
		
	def query (self, *args, **karg):	
		return self.f.query (*args, **karg)
	
	def guess (self, *args, **karg):
		return self.f.guess (*args, **karg)
	
	def close (self):
		self.f.close ()
	
		
if __name__ == "__main__":
	import odbc
	f = Classifier (r"D:\bladepub\search\etc\col\fsc.class")
		
	dbc = odbc.odbc ("pepsi")
	c = dbc.cursor ()
	c.execute ("""
		select top 3 left (egcc, 4) unspsc, 
		isnull (title, '') + ' ' + isnull (seg, '') + ' ' + isnull (fam, '') + ' '+ isnull (cla, '') + ' ' + isnull (com, '') as description
		from iisd.dbo.trainset_unspsc
		where lang = 0 and substring (egcc, 3, 2) <> '00'		
	""")

	matched = 0
	processed = 0
	unclassified = 0
	agreed = 0
	c1 = 0
	c2 = 0
	c3 = 0

	items = c.fetchall ()
	for item in items:	
		processed += 1
		unspsc, title = item
		#res = f.query (title, engine = "id3")
		res = f.guess (title)
		res = res [4:]
		
		if res:
			print(unspsc, res [:2])
			if type (res [0]) == type (()):
				if res [0][0] == unspsc:
					matched += 1
			else:	
				if res [0] == unspsc:
					matched += 1
		else:
			unclassified += 1
	f.close ()
	
	print("_" * 79)
	print((
		"Processed: %d\n"
		"Unclassified: %d\n"
		"Matched: %d\n"
		"Agree: %d\n"
		"1 Class: %d\n"
		"2 Classes: %d\n"
		"3 Classes: %d\n"

		"Accuracy: %2.3f%%\n" % (processed, unclassified, matched, agreed, c1, c2, c3, float (matched) / (processed - unclassified) * 100)
	))

	c.close ()
	dbc.close ()
	


#-----------------------------------------------------------------
# Services
#-----------------------------------------------------------------			
def classify (self, col, query, offset = 0, fetch = 3):
	try:
		classifier = self.searchers [col].classifier
	except:
		return  [404, 0, 0, '']
	
	query = query.strip ()
	if len (query) < 2:
		return  [403, 0, 0, query]
		
	try: tid = threading.currentThread ().getId ()
	except AttributeError: tid = 0
		
	try:	
		result = classifier.guess (self.mpool.get (tid), query, offset, fetch)
	except MemoryError:			
		return self.recover (tid)
		
	self.maintern (tid)
	return result
		
		
		
		



import json
import os
import shutil
import delune
from delune import docqueue
import copy
import threading
from rs4 import logger, pathtool
from delune.bin import indexer
        
OPT_TEMPLATE = {
  "version": 1,
  'name': None,
  'data_dir': [],
  "permission": [],
  'analyzer': {
    "max_terms": 3000,
    "stem_level": 1,
    "strip_html": 0,
    "make_lower_case": 1,
    "ngram": 1,
    "biword": 0,
    "stopwords_case_sensitive": 1,
    "ngram_no_space": 0,    
    "contains_alpha_only": 0,
    "stopwords": [],
    "endwords": [],    
  },
  'indexer': {
    'optimize': 1,
    'force_merge': 0,
    'max_memory': 10000000,
    'max_segments': 10,
    'lazy_merge': (0.3, 0.5),
  },
  'searcher': {
    'max_result': 2000,
    'num_query_cache': 1000
  }
}

def make_config (name, data_dirs = [], version = 1, **kargs):
    config = copy.copy (OPT_TEMPLATE)
    config ["name"] = name
    config ["version"] = version
    if not data_dirs:
        config ["data_dir"] = [name]
    else:
        config ["data_dir"] = []
        if isinstance (data_dirs, str):
            data_dirs = [data_dirs]
        for each in data_dirs:
            config ["data_dir"].append (each)

    for k, v in kargs.items ():
        gotit = False
        for sect in ("analyzer", "indexer", "searcher"):
            if k in config [sect]:
                config [sect][k] = v
                gotit = True
        if not gotit:
            raise NameError ("option `{} = {}` is not recognized".format (k, v))
    return config
      
      
class Documents:
    def __init__ (self, queue, config, model_root, logger):
        self.queue = queue
        self.config = config
        self.logger = logger
        self.model_root = model_root
        self.searcher = None
        self.lock = threading.RLock ()
    
    def __enter__ (self):
        return self
    
    def __exit__ (self, type, value, tb):
        if self.searcher:
            self.searcher.close ()
            self.searcher = None
    
    def __create_searcher (self):
        with self.lock:
            if self.searcher:
                return
            self.queue.si.read ()                
            self.config ["searcher"]["num_query_cache"] = 0 # disable caching
            self.searcher = self.queue.si.get_searcher (**self.config ["searcher"])        
         
    def new (self, id = None, analyzer = None):
        return delune.document (id, analyzer)
    
    def add (self, doc):
        if doc._id:
            self.update (doc)
        else:
            self.queue (0, doc.as_json ())
        
    def update (self, doc):
        assert doc._id, "_id required"
        self.delete (doc._id)
        self.queue (0, doc.as_json ())
    
    def delete (self, id):
        self.queue (1, json.dumps ({"query": {'qs': "_id:" + id}}))
    
    def qdelete (self, qs):
        self.queue (1, json.dumps ({"query": {'qs': qs}}))
    
    def truncate (self, name):
        self.searcher = None
        self.queue.truncate ()
            
    def commit (self):     
        self.searcher = None
        self.queue.commit ()
    
    def index (self):
        indexer.index (self.model_root)
        
    def rollback (self):
        self.queue.rollback ()
    
    def search (self, q, offset = 0, limit = 10, sort = "", snippet = 30, partial = "", nthdoc = 0, lang = "un", analyze = 1, data = 1, qlimit = 1, *arg, **karg):
        self.__create_searcher ()                
        return self.searcher.query (q, offset, limit, sort, snippet, returns, nthdoc, lang, analyze, data, qlimit, *arg, **karg)
        
    def guess (self, q, lang = "en", clf = "naivebayes", top = 0, cond = ""):
        self.__create_searcher ()
        return self.searcher.guess (q, lang, clf, top, cond)
        

class Collection:
    def __init__ (self, name, config, model_root, logger = None):
        self.name = name
        self.config = config
        self.model_root = model_root        
        self.logger = logger
        self.__create_queue ()
        self.documents = Documents (self.queue, config, model_root, logger)
    
    @property
    def config_file (self):
        return os.path.join (self.model_root, "delune", "config", self.name)
    
    @property
    def data_dir (self):
        return [os.path.join (self.model_root, "delune", "collections", each) for each in self.config ['data_dir']]
    
    def close (self):
        if self.dcouments.searcher:
            self.dcouments.searcher.close ()
            
    def __create_queue (self):
        col = delune.collection (
          indexdir = self.data_dir,
          mode = delune.READ,
          analyzer = delune.standard_analyzer (3000, 1, **self.config ['analyzer']),
          version = self.config ['version']
        )
        self.queue = docqueue.DocQueue (col)
    
    def is_active (self):
        return os.path.isfile (self.config_file)
        
    def save (self):
        with open (self.config_file, "w") as f:
            f.write (json.dumps (self.config))
    
    def drop (self, include_data = False):
        if self.documents.searcher:
            self.documents.searcher.close ()
            self.documents.searcher = None
            
        if os.path.isfile (self.config_file):
            os.remove (self.config_file)
            
        if include_data:
            for d in self.data_dir:
                shutil.rmtree (d)
    
    def undrop (self):
        raise SystemError ("cannot undrop")
    

class Delune:
    def __init__ (self, model_root):
        self.model_root = model_root
        self.logger = logger.screen_logger ()
        delune.configure (1, self.logger)
        pathtool.mkdir (self.getdir ("config"))
        pathtool.mkdir (self.getdir ("collections"))
    
    def getdir (self, *d):
        return os.path.join (self.model_root, "delune", *d)
        
    def lscol (self):
        cols = []
        for each in os.listdir (self.getdir ("config")):
            cols.append (each)
        return cols
        
    def load (self, name):
        if name not in self.lscol ():
            raise NameError ("collection not found")        
        with open (os.path.join (self.getdir ("config", name))) as f:
            config = json.loads (f.read ())            
        return Collection (name, config, self.model_root)
    
    def create (self, name, data_dirs = [], version = 1, **kargs):
        config = make_config (name, data_dirs, version, **kargs)        
        return Collection (name, config, self.model_root, self.logger)

    


import types
from . import querylexer

PLUS = '+'
MINUS = '-'
TIMES = '*'
DIVIDE = '/'
LPAREN = '('
RPAREN = ')'
EOF = 'EOF'

class ArithmeticParser:
	"""Parses a token list and returns a nested expression structure in
	prefix form."""
	def __init__(self, tokens):
		self.tokens = tokens


	def peek_token(self):
		"""Looks at the next token in our token stream."""
		return self.tokens[0]
		
	def consume_token(self):
		"""Pops off the next token in our token stream."""
		next_token = self.tokens[0]
		del self.tokens[0]
		return next_token


	def parse_expression(self):
		"""Returns a parsed expression.  An expression may consist of a
		bunch of chained factors."""
		expression = self.parse_factor()
		while True:
			if self.peek_token() in (PLUS, MINUS):
				operation = self.consume_token()
				factor = self.parse_factor()
				expression= [operation, expression, factor]
			else: break

		return expression


	def parse_factor(self):
		"""Returns a parsed factor.	A factor may consist of a bunch of
		chained terms."""
		factor = self.parse_term()
		while True:
			if self.peek_token() in (TIMES, DIVIDE):
				operation = self.consume_token()
				term = self.parse_term()
				factor = [operation, factor, term]
			else: break
		return factor

	def parse_term(self):
		"""Returns a parsed term.  A term may consist of a number, or a
		parenthesized expression."""
		if self.peek_token() not in (LPAREN,):
			return str(self.consume_token())
		else:
			self.consume_token() ## eat up the lparen
			expression = self.parse_expression()
			self.consume_token() ## eat up the rparen
			return expression



def main():
	"""Some test code."""
	a= '((alltitle:"medical^3 service" -monitoring) or ("medical^3 service" -monitoring) or (alltitle:"service^3 medical" -monitoring) or ("service^3 medical" -monitoring))'
	b= '(alltitle:"medical^3 service" -monitoring) file:"adada sada" -file:"qwe-qwe-q"'
	c= '((ground range:2123) or (service)) range:2342'
	tokens = querylexer.QueryLexer (b).tokenize ()

	#print tokens	
	expression = ArithmeticParser(tokens).parse_expression()
	print(expression)


if __name__ == '__main__':
	main()



from rs4 import pathtool
from delune import _delune
import os
import glob

class FileSystem:
	def __init__ (self, paths):
		self.paths = []
		if type (paths) is type (""):
			paths = [paths]
		for path in paths:
			path = os.path.normpath (path)
			self.paths.append (path)
			pathtool.mkdir (path)
				
		self.dirs = {}
		
		self.master = self.paths [0]		
		self.__cache = {}
		self.__cacheDir ()
		self.__dirty = 0

	def getpathes (self):
		pathes = {}
		i = 0
		for path in self.paths:
			pathes [i] = path
			i += 1
		return pathes

	def notify (self):		
		self.__dirty = 1

	def getmaster (self):
		return self.master

	def __cacheDir (self):
		self.__cache = {}		
		for path in self.paths:
			self.dirs [path] = 0
			segs = {}
			for file in os.listdir (path):
				try: segs [int (file [:-4])] = None
				except ValueError: pass	
			
			for seg in segs:
				size = 0
				for segfile in glob.glob (os.path.join (path, str (seg) + ".*")):
					size += os.stat (segfile).st_size
				
				self.__cache [seg] = (path, size)
				self.dirs [path] += size
		
		self.__dirty = 0
		
	def dirty (self):
		self.__dirty = 1

	def get (self, seg):
		if self.__dirty:
			self.__cacheDir ()

		try:
			return self.__cache [int (seg)][0]
		except KeyError:
			self.__cacheDir ()
			try: return self.__cache [int (seg)][0]
			except: return None

	def new (self, seg):
		prev = self.get (seg) #previously failed file
		if prev: return prev
		
		if self.__dirty:
			self.__cacheDir ()
		
		filesizes = list(self.dirs.items ())
		filesizes.sort (key = lambda x: x[1])
		
		freespaces = {}		
		for path in list(self.dirs.keys ()):
			freespaces [path] = _delune.getdiskspace (path)
		
		for path, size in filesizes:
			if freespaces [path] > 2000:
				return path
		
		diskspaces = list(self.dirs.items ())
		diskspaces.sort (key = lambda x: x[1], reverse = True)		
		
		return diskspaces [0][0]

	def getpathindex (self, seg):
		return self.paths.index (self.get (int (seg)))
	
	def getsegmentpath (self, seg):
		return self.paths [self.getpathindex (seg)]
	
	def hasDeleted (self, seg):
		return os.path.isfile (os.path.join (self.getsegmentpath (seg), "%s.del" % seg))
			
	def getsegmentsize (self, seg):
		path = self.getsegmentpath (seg)
		info = {}
		for file in glob.glob (os.path.join (path, str (seg) + ".*")):
			info [os.path.split (file) [-1][-3:]] = os.stat (file).st_size
		return info
		
	def dirinfo (self):
		if self.__dirty:
			self.__cacheDir ()		
		n = {}
		for k, v in list(self.__cache.items ()):
			n [str (k)] = v
		return n
		
	def all (self):
		if self.__dirty:
			self.__cacheDir ()

		return list(self.__cache.keys ())
	
	def segments (self):
		return self.all ()
		
	def remove (self, seg):
		if self.__dirty:
			self.__cacheDir ()

		for file in glob.glob (os.path.join (self.get (seg), str (seg) + ".*")):
			try:
				os.remove (file)
			except OSError as why:
				if why [0] != 2:
					raise
			
		self.__dirty = 1


def get_segment_files (paths):
	files = {}
	for path in paths:
		segs = {}
		for file in os.listdir (path):
			try: segs [int (file [:-4])] = None
			except ValueError: pass
								
		for seg in segs:
			files [seg] = []
			for segfile in glob.glob (os.path.join (path, str (seg) + ".*")):
				files [seg].append ((segfile, os.stat (segfile).st_size, os.stat (segfile).st_mtime))
		
		segments = os.path.join (path, "segments")
		if os.path.isfile (segments):
			files ['segments'] = (segments, os.stat (segments).st_size, os.stat (segments).st_mtime)
		
	return files


if __name__ == "__main__":
	f = FileSystem (
		[
			"d:/bladedb/0/bladeindex/ontowns",
			"d:/bladedb/1/bladeindex/ontowns",
			"d:/bladedb/2/bladeindex/ontowns",
		],
		None
	)
	
	
	print(f.getsegmentpath (1514))
	print(f.getsegmentsize (1514))
	
	


import delune
import os
import json
from rs4 import pathtool
import json
import codecs
import time
import shutil
from rs4 import pathtool

RESOURCE_DIR = None

def getdir (*d):
	global RESOURCE_DIR
	return os.path.join (RESOURCE_DIR, "delune", *d)
	
def load_data (alias, numthreads, plock):
	def normpath (path):
		if os.name == "nt":
			return path.replace ("/", "\\")
		return path.replace ("\\", "/")

	with codecs.open (getdir ("config", alias), "r", "utf8") as f:
		colopt = json.loads (f.read ())			
		colopt ['data_dir'] = [getdir ("collections", normpath(d)) for d in colopt ['data_dir']]
	
	name = "standard"
	if "name" in colopt ["analyzer"]:
		name = colopt ["analyzer"].get ("name")			
		del colopt ["analyzer"]["name"]
	analyzer_class = delune.get_analyzer (name)
	
	if 'classifier' in colopt:		
		analyzer = analyzer_class (10000, numthreads, **colopt ["analyzer"])
		col = delune.model (colopt ["data_dir"], delune.READ, analyzer, plock = plock, version = colopt.get ("version", 1))
		actor = col.get_classifier (**colopt.get ('classifier', {}))
	else:
		analyzer = analyzer_class (8, numthreads, **colopt ["analyzer"])
		col = delune.collection	(colopt ["data_dir"], delune.READ, analyzer, plock = plock, version = colopt.get ("version", 1))	
		actor = col.get_searcher (**colopt.get ('searcher', {}))
		actor.create_queue ()
	delune.assign (alias, actor)

#------------------------------------------------------------------

def decorate (app):
	global RESOURCE_DIR
	RESOURCE_DIR = app.config.resource_dir
	
	@app.route ("/<alias>/config", methods = ["GET"])
	@app.permission_required (["index", "replica"])
	def config (was, alias):
		fn = getdir ("config", alias)
		return was.response.file (fn, "application/json")
	
	@app.route ("/list", methods = ["GET"])
	@app.permission_required (["replica", "index"])
	def collections (was):
		return was.response.API ({'collections': list (delune.status ().keys ())})	
	
	@app.route ("/<alias>", methods = ["GET", "POST", "PATCH", "DELETE", "OPTIONS"])
	@app.permission_required (["replica", "index"])
	def collection (was, alias, side_effect = ""):
		fn = getdir ("config", alias)
		if was.request.command == "get":
			if side_effect == "undo":
				for mark in "#-":
					if os.path.isfile (getdir ("config", mark + alias)):
						os.rename (
							getdir ("config", mark + alias),
							getdir ("config", alias)
						)
						was.setlu ("delune:collection")					
						return was.response ("201 Accept", was.response.API ())
				return was.response.Fault ("404 Not Found", 20100, "resource already commited")
			
			if not delune.get (alias):
				return was.response.Fault ("404 Not Found", 40401, "resource %s not exist" % alias)
				
			status = delune.status (alias)
			conf = getdir ("config", alias)
			with codecs.open (conf, "r", "utf8") as f:
				colopt = json.loads (f.read ())		
				status ['colopt'] = {
					'data': colopt,
					'mtime': 	os.path.getmtime (conf),
					'size': 	os.path.getsize (conf),
					'path': conf
				}
			return was.response.API (status)
				
		if was.request.command == "delete":
			if not os.path.isfile (fn):
				return was.response.Fault ("404 Not Found", 40401, "resource not exist")
			
			a, b = os.path.split (fn)
			if side_effect == "data":
				newfn = os.path.join (a, "-" + b)
			else:
				newfn = os.path.join (a, "#" + b)		
			os.rename (fn, newfn)
			was.setlu ("delune:collection")
			return was.response.API ()
		
		for mark in "#-":
			if os.path.isfile (getdir ("config", mark + alias)):			
				return was.response.Fault ("406 Conflict", 40601, "removed resource is already exists, use UNDO")
				
		if was.request.command == "post" and delune.get (alias):
			return was.response.Fault ("406 Conflict", 40601, "resource already exists")		
		elif was.request.command == "patch" and not delune.get (alias):
			return was.response.Fault ("404 Not Found", 40401, "resource not exist")
				
		with codecs.open (fn, "w", "utf8") as f:
			f.write (was.request.body.decode ("utf8"))
		
		was.setlu ("delune:collection")		
		return was.response.API ()

	# replica -------------------------------------------------------
			
	@app.route ("/<alias>/locks", methods = ["GET"])
	@app.permission_required ("replica")
	def locks (was, alias):	
		return was.response.API ({"locks": delune.get (alias).si.lock.locks ()})

	@app.route ("/<alias>/locks/<name>", methods = ["POST", "DELETE", "OPTIONS"])
	@app.permission_required ("replica")
	def handle_lock (was, alias, name):	
		if was.request.command == "post":
			delune.get (alias).si.lock.lock (name)		
			return was.response.API ()
		delune.get (alias).si.lock.unlock (name)
		return was.response.API ()
	
	@app.route ("/<alias>/groups/<group>/<fn>", methods = ["GET"])
	@app.permission_required ("replica")
	def getfile (was, alias, group, fn):
		s = delune.status (alias)
		if group == "primary":
			path = os.path.join (s ["indexdirs"][0], fn)
		else:
			path = os.path.join (s ["indexdirs"][0], group, fn)
		return was.response.file (path)

	@app.route ("/<alias>/groups/<group>/segments/<fn>", methods = ["GET"])
	@app.permission_required ("replica")
	def getsegfile (was, alias, group, fn):
		s = delune.status (alias)
		seg = fn.split (".") [0]
		if group == "primary":
			if seg not in s ["segmentsizes"]:
				return was.response.Fault ("404 Not Found", 40401, "resource not exist")
			path = os.path.join (s ["segmentsizes"][seg][0], fn)	
		else:
			path = os.path.join (s ["indexdirs"][0], group, fn)
		return was.response.file (path)
	
	# index ------------------------------------------------------------

	@app.route ("/<alias>/commit", methods = ["GET"])
	@app.permission_required ("index")
	def commit (was, alias):
		delune.get (alias).queue.commit ()
		return was.response.API ()

	@app.route ("/<alias>/rollback", methods = ["GET"])
	@app.permission_required ("index")
	def rollback (was, alias):
		delune.get (alias).queue.rollback ()
		return was.response.API ()
	
	@app.route ("/<alias>/documents", methods = ["POST", "DELETE", "OPTIONS"])
	@app.permission_required ("index")
	def documents (was, alias, truncate_confirm = "", q = None, lang = "en", analyze = 1):
		if was.request.command == "delete":
			if q:
				delune.get (alias).queue (1, json.dumps ({"query": {'qs': q, 'lang': lang, 'analyze': analyze}}))
				return was.response.API ()
			elif truncate_confirm != alias:
				return was.response.Fault ("400 Bad Request", 40003, 'parameter truncate_confirm=(alias name) required')			
			delune.get (alias).queue.truncate ()
			return was.response.API ()
	
		delune.get (alias).queue (0, was.request.body)
		return was.response.API ()
	
	@app.route ("/<alias>/documents/<_id>", methods = ["GET", "DELETE", "PUT", "PATCH", "OPTIONS"])
	@app.permission_required ("index")
	def document (was, alias, _id, nthdoc = 0):
		if was.request.command == "get":		
			return was.response.API (delune.query (alias, "_id:" + _id, nthdoc = nthdoc))
		
		delune.get (alias).queue (1, json.dumps ({"query": {'qs': "_id:" + _id}}))
		if was.request.command in ("patch", "put"):
			delune.get (alias).queue (0, was.request.body)
		return was.response.API ()




import requests
import os
import json
import delune
from rs4 import pathtool
import time
import socket
import pickle
from rs4.psutil import processutil
import sys, getopt
from rs4.psutil import service, daemon_class
from . import indexer

DUPLICATE_LOCK = "replicate-%s" % socket.getfqdn ()
DEBUG = True

def normpath (d):
	d = d.replace ("\\", "/")
	if DEBUG:
		d = d.replace ('D:/apps/', 'd:/var/')
	return d

class	Replicator (indexer.Indexer):
	NAME = "replicator"

	def __init__ (self, logpath, varpath, consol, config):
		self.config_dir = None
		indexer.Indexer.__init__ (self, logpath, varpath, consol, config)
		
	def download (self, alias, session, item):
		origin = self.config ["origin"]
		
		remotepath, savepath	= item
		pathtool.mkdir (os.path.split (savepath)[0])
		surl = "ops/" + remotepath
		url = "%s/%s" % (origin, surl)
		self.log ("downloading from %s" % url, "info", alias)
		
		response = session.get(url, stream=True)
		cl = int (response.headers ["Content-Length"])						
		rl = 0
		
		if not response.ok:
			self.log ("download failed  HTTP %d: %s" % (response.status_code, url), 'fail', alias)
			return -1
			
		with open(savepath, 'wb') as handle:
			if cl:
				for block in response.iter_content(1024):
					if self.enter_shutdown:
						raise daemon_class.ExitNow
					rl += len (block)
					if rl % 10240000 == 0:
						self.log ("%d%% downloaded %s" % (rl/cl * 100, surl), "info", alias)
					handle.write(block)				
				self.log ("downloaded %d of %d (%d%%) from %s" % (rl, cl, rl/cl * 100, surl), alias)
				
		if rl != cl:
			self.log ("incomplete download: %s" % url, 'fail', alias)
			return -1
		return 1			
	
	def replicate_collection (self, alias, remote_session, remote_status):
		origin = self.config ["origin"]
		keep_going = True		
		for name, ctime in remote_status ["locks"]:
			if name == "index":
				self.log ("origin %s is under indexing, skip updating index...", "warn", alias)
				return
		
		if keep_going:
			local_segments = remote_status ["segments"][0]
			if not os.path.isfile (local_segments):
				local_latest_segments = -1				
			else:
				with open (local_segments, "rb") as f:
					temp = pickle.load (f)
				local_latest_segments = max (list(temp.segmentInfo.keys ()))
			origin_segments = [seginfo [0] for seginfo in remote_status ['segmentinfos']]
			
			if not origin_segments:
				self.log ("origin %s is empty", "warn", alias)
				keep_going = False
			elif max (origin_segments) < local_latest_segments:
				self.log ("origin is older version, maybe next time...", "warn", alias)
				keep_going = False
		
		que = []
		new_file = False		
		if keep_going:
			for d in remote_status ['indexdirs']:
				pathtool.mkdir (normpath (d))
			
			for group in remote_status ["segmentfiles"]:
				segs = remote_status ["segmentfiles"][group]
				
				if 'segments' in segs:
					sfn = segs.pop ('segments')[0]
					dfn = normpath (sfn)
					if group == "primary":
						que.append (("{}/groups/{}/segments".format (alias, group), dfn + ".new"))
					else:
						que.insert (0, ("{}/groups/{}/segments".format (alias, group), dfn))
				
				for seg, files in segs.items ():
					if group == "primary" and int (seg) not in origin_segments:
						continue
					
					for sfn, ssize, smtime in files:
						dfn = normpath (sfn)
						if not os.path.isfile (dfn):
							que.insert (0, ("{}/groups/{}/segments/{}".format (alias, group, str(seg) + sfn [-4:]), dfn))
							new_file = True
						else:	
							dmtime = os.path.getmtime (dfn)
							dsize = os.path.getsize (dfn)
							if dsize == ssize and smtime <= dmtime:
								continue
							que.insert (0, ("{}/groups/{}/segments/{}".format (alias, group, str (seg) + sfn [-4:]), dfn))
							new_file = True
		
		if new_file:
			self.log ("downliad segments", "echo", alias)
			failed = 0
			for item in que:				
				#col, entity, group, fn, savepath
				if self.download (alias, remote_session, item) == -1:
					failed += 1
			
			local_segments = remote_status ["segments"][0]
			if failed:
				self.log ("replication failed, maybe mext time", "fail", alias)
				os.remove (local_segments + ".new")
			else:
				if os.path.isfile (local_segments):
					os.remove (local_segments)	
				os.rename (local_segments + ".new", local_segments)			
		
		colopt = remote_status ["colopt"]
		dcolopt = normpath (colopt ['path'])
		if not os.path.isfile (dcolopt) or colopt ["mtime"] > os.path.getmtime (dcolopt) or colopt ["size"] != os.path.getsize (dcolopt):
			with open (dcolopt, "w") as out:
				out.write (json.dumps (colopt ['data']))			
			self.log ("update collection configure", "info", alias)
		
	def replicate (self):
		remote_session = requests.Session ()
		origin = self.config ["origin"]
		
		while 1:
			alias = self.q.get ()
			if not alias:
				return
			if self.enter_shutdown:
				return
			with self.lock: self.working [alias] = None
			self.log ("checking...", "echo", alias)
			local_col = None
			try:
				remote_session.post ("%s/ops/%s/locks/%s" % (origin, alias, DUPLICATE_LOCK))
				r = remote_session.get ("%s/ops/%s" % (origin, alias))
				remote_status = r.json ()
				if self.config_dir is None:
					self.config_dir = os.path.dirname (remote_status ["colopt"]["path"])
					pathtool.mkdir (self.config_dir)
					
				local_col = delune.collection (indexdir = [normpath (d) for d in remote_status ['indexdirs']], mode = delune.MODIFY)
				if local_col.lock.isplocked ("replicate"):
					self.log ("on replicating for %s, maybe next", "warn", alias)
					continue
				local_col.lock.lock ("index", str (os.getpid ()))		
				self.replicate_collection(alias, remote_session, remote_status)				
				local_col.removeDeletables ()
			
			except Exception as why:
				if not isinstance (why, daemon_class.ExitNow):
					self.trace ()

			try:
				remote_session.delete ("%s/ops/%s/locks/%s" % (origin, alias, DUPLICATE_LOCK))
			except:
				self.trace ()
			
			local_col and local_col.lock.unlock ("index")
			with self.lock: del self.working [alias]
			if self.enter_shutdown: return
	
	def maintern (self, origin_side):
		if not origin_side:
			# possibly error
			return
		if not self.config_dir:
			return 
		for alias in os.listdir (self.config_dir):
			if alias not in origin_side:
				if alias [0] in "#-":
					continue
				os.rename (os.path.join (self.config_dir, alias), os.path.join (self.config_dir, "-" + alias))
				self.log ("remote collection has been removed", 'warn', alias)
		indexer.Indexer.maintern (self, self.config_dir)
						
	def run (self):
		self.threading (self.replicate)
		while 1:			
			try:
				origin_side = requests.get ("%s/ops/list" % self.config ["origin"]).json ()["collections"]
			except:
				self.trace ()
			else:
				self.maintern (origin_side) 
				self.reque (origin_side)
						
			if not self.wait_for_interval ():
				break			
							
	 	
def usage ():
	print ("""usage: 
  delune replicate [options] [start|stop|status|restart]

options:
  -d: daemonize
  -e, --exclude: <alias,...>
  -o, --origin: <origin>  
  -i, --interval: <int> interval seconds 
  -t, --threads: <int>
  
example:
  delune replicate -i 300 -vo http://192.168.1.200:5000/v1
	""")
	sys.exit ()

VARPATH = "/var/tmp/delune/replicator"
LOGPATH = "/var/log/delune/replicator"
	 
def main ():
	cmd, config, console = indexer.parse_argv ("dv:o:t:i:e:", ["exclude=", "origin=", "threads=", "interval=", "resource-dir", "help"], usage)
	if not cmd or cmd in ("start", "restart"):
		assert "origin" in config, "origin required"
	assert "resource_dir" not in config, "invalid option: resource_dir" 
		
	pathtool.mkdir (VARPATH)
	servicer = service.Service ("delune/{}".format (Replicator.NAME), VARPATH)
	if cmd and not servicer.execute (cmd):
		return
	if console and servicer.status (False):
		raise SystemError ("daemon is running")
	
	s = daemon_class.make_service (Replicator, LOGPATH, VARPATH, console, config)
	s.start ()
	s.close ()


if __name__ == "__main__":
	main ()	


from delune import indexinfo
from rs4 import confparse, logger
import memory
from . import searcher
from . import query

class Searcher:
	def __init__ (self, path, debug = 0):
		self.path = path
		query.Query.DEBUG = debug
		memory.create (1, 32768, 128, "segment", logger.screen_logger ())
		inf = indexinfo.IndexInfo (confparse.ConfParse (path))
		self.s = searcher.Searcher (inf)
		self.s.initialize ()

	def query (self, q, offset = 0, fetch = 0, *args, **karg):
		return self.s.query (q, offset, fetch, *args, **karg)


if __name__ == "__main__":
	f = Searcher (r"D:\bladepub\proto\etc\col\unifiedbid")
	d = f.query ("service", 0, 10, summary = 30)
	
	print(d [2], d [4])
	
	
	
	


from . import segment
from delune import _delune
import os

class Term:
	def __init__ (self, term, fdno, df, frqPosition, prxPosition, skipDelta, prxLength):
		self.term = term
		self.fdno = fdno
		self.df = df
		self.frqPosition = frqPosition		
		self.prxPosition = prxPosition		
		self.skipDelta = skipDelta		
		self.prxLength = prxLength		
	
	def __lt__ (self, other):
		return self.fdno == other.fdno and self.term < other.term or self.fdno < other.fdno
	
	def __eq__ (self, other):
		return self.fdno == other.fdno and self.term == other.term
		
	def __cmp__ (self, other):
		if self.fdno == other.fdno:
			return cmp (self.term, other.term)	
		else:
			return cmp (self.fdno, other.fdno)	
		
	def __repr__ (self):
		return "<Term %s: %d>" % (self.term, self.fdno)
			

class SegmentReader (segment.Segment):
	def __init__ (self, si, seg):
		self.si = si
		segment.Segment.__init__ (self, si.fs.get (seg), seg, 'r', self.si.plock, version = self.si.version)
	
	def readPosting (self, mem, df, doff, poff, skip, plen, zoff = -1, zlen = -1, getprox = 0):
		return self.tf.read (mem, df, doff, poff, skip, plen, zoff, zlen, getprox)
	
	def getNumDoc (self):
		return self.numDoc
	
	def getSortMapPointer (self, fdno):
		if self.si.getType (fdno) is None: 
			return -1
		
		self.si.getSortMapFields () # create cache
		
		pointer = 0		
		i = 0
		while 1:
			n, s = self.si.cache_sortmap [i]
			if n == fdno: break
			pointer += self.numDoc * s
			i += 1
		
		return pointer, s
		
	def getDeletedCount (self):
		if self.rd: return self.rd.count ()
		else: return 0
			
	def getBits (self):
		return self.rd and self.rd.getbits () or None
			
	def getDocument (self, mem, docId, summary = 0, terms = None, nthdoc = 0):
		return self.fd.read (mem, docId, summary, terms or [], nthdoc)		
		
	def getTermInfo (self, mem, term, fdno):
		return self.ti.get (mem, term [:100], fdno)
	
	def getTermInfos (self, mem, term, fdno):
		return self.ti.get_wildcard (mem, term [:100], fdno)		
	
	def getTermInfos2 (self, mem, terms, fdno):
		return self.ti.get_range (mem, terms [0], terms [1], fdno)
				
	def delete (self, docid):
		if self.rd is None:
			self.create_bitvector ()
		else:
			self.reload_deleted ()
		self.rd.set (docid)
		
	
class SegmentBulkReader (SegmentReader):
	def __init__ (self, si, seg):
		self.si = si
		segment.Segment.__init__ (self, si.fs.get (seg), seg, 'm', version = self.si.version)
	
	def getMergeInfo (self):
		return ( 
			self.numDoc, 
			self.fs_tfq,
			self.fs_prx, 
			self.fs_fdi, 
			self.fs_fda, 
			self.fs_smp,
			self.rd and self.rd.getbits () or None
		)	
	
	def advanceTermInfo (self):		
		ti = self.ti.advance ()
		return Term (*ti)
		


from delune.searcher import document
from rs4 import unistr

class LabeledDocument (document.Document):
	def __init__ (self, label, text, lang = "un", encoding = None):
		self.fields = {}
		self.label = unistr.makes (label)
		self.text = unistr.makes (text, encoding)
		self.lang = lang
	
	def add_field (self, name, value, ftype = "Text", lang = "un", encoding = None, option = {}):
		if name in ("label", "default"):
			raise NameError("Field Name `label' or `default' is not allowed")
		document.Document.add_field (self, name, value, ftype, option, encoding, lang)




import delune
import os
import json

#-----------------------------------------------------------------

def decorate (app):
    
    @app.route ("/test", methods = ["GET"])
    def test (was):
        return was.response.API (tested = "DELUNE PUBLIC API")
    
    @app.route ("/<alias>/documents/<_id>", methods = ["GET"])
    def documents (was, alias, _id, nthdoc = 0):
        return was.response.API (delune.query (alias, "_id:" + _id, nthdoc = nthdoc))
            
    @app.route ("/<alias>/search", methods = ["GET", "POST", "OPTIONS"])
    def query (was, alias, **args):
        q = args.get ("q")
        if not q:
            return was.response.Fault ("400 Bad Request", 40003, 'parameter q required')
        
        l = args.get ("lang", "en")
        analyze = args.get ("analyze", 1)
        
        o = args.get ("offset", 0)
        f = args.get ("limit", 10)
        s = args.get ("sort", "")
        w = args.get ("snippet", 30)    
        r = args.get ("partial", "")
        d = args.get ("nthdoc", 0)    
        data = args.get ("data", 1)
        
        if type (q) is list:
            return was.response.API ([delune.query (alias, eq, o, f, s, w, r, l, d, analyze, data, limit = 1) for eq in q])
        return was.response.API (delune.query (alias, q, o, f, s, w, r, d, l, analyze, data, limit = 1))

    @app.route ("/<alias>/guess", methods = ["GET", "POST", "OPTIONS"])
    def guess (was, alias, **args):    
        # args: q = '', l = 'en', c = "naivebayes", top = 0, cond = ""
        q = args.get ("q")
        if not q:
            return was.response.Fault ("400 Bad Request", 40003, 'parameter q required')                
        l = args.get ("lang", 'en')
        c = args.get ("clf", 'naivebayes')
        top = args.get ("top", 0)
        cond = args.get ("cond", '')
        if type (q) is list:
            return was.response.API ([delune.guess (alias, eq, l, c, top, cond) for eq in q])
        return was.response.API (delune.guess (alias, q, l, c, top, cond))
        
    @app.route ("/<alias>/cluster", methods = ["GET", "POST", "OPTIONS"])
    def cluster (was, alias, **args):
        q = args.get ("q")
        if not q:
            return was.response.Fault ("400 Bad Request", 40003, 'parameter q required')
        l = args.get ("lang", 'en')    
        if type (q) is list:
            return was.response.API ([delune.cluster (alias, eq, l) for eq in q])    
        return was.response.API (delune.cluster (alias, q, l))

    @app.route ("/<alias>/stem", methods = ["GET", "POST", "OPTIONS"])
    def stem (was, alias, **args):
        q = args.get ("q")
        if not q:
            returnwas.response.Fault ("400 Bad Request", 40003, 'parameter q required')
        if isinstance (q, str):
            q = q.split (",")
        l = args.get ("lang", 'en')
        return was.response.API (dict ([(eq, " ".join (delune.stem (alias, eq, l))) for eq in q]))    

    @app.route ("/<alias>/analyze", methods = ["GET", "POST", "OPTIONS"])
    def analyze (was, alias, **args):
        q = args.get ("q")
        if not q:
            return was.response.Fault ("400 Bad Request", 40003, 'parameter q required')
        l = args.get ("lang", 'en')
        return was.response.API (delune.analyze (alias, q, l))



import pickle as pk
import os

def digest (field, word, version = "1.0"):
	return _delune.wint4 (hash (field + ":" + word))

def serialize (d):
	return pk.dumps (d, 1)

def deserialize (d):	
	return pk.loads (d)

def check_config (path):
	from rs4 import confparse
	this = confparse.ConfParse (path)
	confdir, alias = os.path.split (path)
	for file in os.listdir (confdir):
		if file == alias: continue
		other = confparse.ConfParse (os.path.join (confdir, file))
		for this_home in this.getopt ("home"):
			try:
				other.getopt ("home").index (this_home)
			except:
				continue
			else:
				raise LookupError("Directory `%s' is in used by `%s' index" % (this_home, file))



import math
from delune import _delune

class Classifier:
	def __init__ (self, reader, analyzer, featset, use_top_features, logger = None):
		self.reader = reader
		self.analyzer = analyzer
		self.featset = featset
		self.logger = logger
		self.numfeat = len (self.featset)
		self.use_top_features = use_top_features		
		self.verbose = False
		self.setopt ()
			
	def setopt (self, **opts):
		self.verbose = opts.get ("verbose", self.verbose)
	
	def getopt (self, name):
		pass
				
	def match (self, name):
		return name in ("nb", "bayes", "naivebayes")
	
	def close (self):
		pass
	
	def getCoWord (self, mem, t1, t2):
		#(findex, co-freq)
		k = self.reader.getCoOccurrence (mem, t1, t2)
		if k is None:
			return 0
		return k [1]	
				
	def getFeatures (self, mem, qs, lang):
		if type (qs) is list:
			return qs
		terms = self.analyzer.term (qs, lang)		
		#print ([term for term in terms if term [0] in self.featset])
		if not terms: return []
		return self.reader.getFeatures (
			mem, 
			[term for term in terms if term [0] in self.featset], 
			self.numfeat, 
			self.use_top_features
			)
			
	def guess (self, mem, qs, lang = "un", cond = ""):
		terms = self.getFeatures (mem, qs, lang)
		if not terms: return []
		
		classifier = _delune.Classifier (mem, self.reader.corpus, self.numfeat, self.reader.getN ())
		for term, tf in terms:
			if self.reader.readTermInfo (mem, term):
				classifier.load ()
				has_prob = classifier.bayes (tf)
				if has_prob == 0:
					# all zero probability, abort
					return []

		result = classifier.get (0, self.reader.numpools)
		classifier.close ()
		return self.reader.translate (result)
	
	


import os
import sys
import types
from delune import _delune, memory
from delune.searcher import indexer, document
from ..util import util
from .segment import composedsegmentwriter
from delune.searcher.segment import segment as document_segment
import math
import delune
import pickle

class ExitNow (Exception): 
	pass


class PoolInfo:
	def __init__ (self, name, df):
		self.name = name
		self.df = df
		self.tf = 0

	def __lt__ (a, b):
		return a.name < b.name

	def __repr__ (self):
		return self.name


class Corpus:
	def __init__ (self):
		self.N = 0
		self.pools = []
		self.numpools = 0
		self.poolno = {}
		self.terms = {}
		self.scores = []
		self.features = {}
		
			
class Trainer (indexer.Indexer):
	def __init__ (self, si):
		indexer.Indexer.__init__ (self, si)		
		self.indexer = None
		self.corpus = Corpus ()
		
	def init (self):
		self.si.log ("initializing trainer...", "info")
		if not self.isIndexable ():
			raise ExitNow("`%s' not Indexable" % self.si.getAlias ())
		self.writer = composedsegmentwriter.ComposedSegmentWriter (self.si)
		self.corpus.N = self.writer.getIndexedNumDoc ()
	
	def rebuildIndex (self, breader):		
		self.si.log ("rebuild index for shrinking...", "info")
		bog = self.corpus.terms
		col = delune.collection (os.path.join (self.si.fs.getmaster (), "%s.ts" % self.si.alias), delune.APPEND, logger = self.si.logger, version = self.si.version)
		old_segment = col.getLastSegment ()
		new_path, new_segment = col.clone (old_segment)
		
		mem = memory.Memory (1, 1024, 10).get ()
		modifier = document_segment.DocumentSegment (new_path, new_segment, "w")		
		t = breader.getNumDoc ()
		i = 0
				
		while i < t:
			field, summary = breader.getDocument (mem, i)
			label, terms = util.deserialize (field)
			
			fterms = []
			for term, tf in terms:
				try:
					termid = bog [term]
				except KeyError:
					pass
				else:
					fterms.append ((termid, tf))					
			
			fterms.sort ()
			lterms = []
			for fid, tf in fterms:
				lterms.extend ([fid, tf])
				
			modifier.add_content ((label, lterms), summary)
			i += 1
		
		modifier.close ()
		col.removeSegment (old_segment)
		col.addSegment (new_segment, i)
		col.flush ()
		self.si.log ("rebuild done", "info")
					
	def add_document (self, labeledDocument):
		# labeleddocument		
		if self.indexer is None:
			if "optimized" in self.si.segments.parameters:
				if not self.si.rebuild:
					raise AssertionError ("Already optimized, can't append")
			
			col = delune.collection (os.path.join (self.si.fs.getmaster (), "%s.ts" % self.si.alias), self.si.rebuild and delune.CREATE or delune.WRITE, self.si.analyzer, self.si.logger, version = self.si.version)
			self.indexer = col.get_indexer (force_merge = True)
			
		terms = self.si.analyzer.term (labeledDocument.text)
		if not terms: return
		
		doc = document.Document ()
		doc.field ("label", labeledDocument.label, delune.STRING)
		doc.field ("default", labeledDocument.text, delune.TEXT, lang = labeledDocument.lang )
		for name, (val, ftype, lang, option) in list(labeledDocument.fields.items ()):
			doc.field (name, val, ftype, option, lang = lang)
		doc.document ([labeledDocument.label, terms])
		self.indexer.add_document (doc)
		
	def close (self):
		if self.indexer:
			self.closeIndexer ()
					
		if self.writer:
			self.writer.flush ()
			self.writer.close ()
		self.si.lock.unlock ("index")
		self.si.close ()
		self.si = None
		
	def scoringFeaures (self, selector, select_way, prune_df_min, prune_df_max):
		terms = self.corpus.terms	
		sl = _delune.Selector (len (self.corpus.pools), self.corpus.N, len (terms))
		
		if select_way == delune.SUM:		method = 0
		elif select_way == delune.MAX:	method = 1
		elif select_way == delune.AVG:	method = 2
		
		selector = getattr (sl, selector)		
		self.si.log ("setting corpus")
		fi = []
		for pool in self.corpus.pools:
			fi.append ((pool.df, pool.tf))
		sl.addCorpus (fi)
		
		self.si.log ("calculating feature's score...")
		scores = []
		t = len (terms)
		c = 0
		p = 0
		for term in terms:			
			c += 1
			if c % 1000 == 0:
				self.si.log ("calculating feature %d/%d (%d pruned)" % (c, t, p))
				
			tdf, fi = self.writer.featureClassInfo (term, self.corpus)
			if not tdf: continue
			if prune_df_min > 0 and tdf < prune_df_min:
				p += 1
				continue
			if prune_df_max > 0 and tdf > prune_df_max:
				p += 1
				continue			
			sl.add (fi)			
			score = selector (method)
			scores.append ((term, score))			

		self.si.log ("sorting by score")
		scores.sort (key = lambda x: x [1], reverse = True)
		self.si.log ("feature selection, done.")
		sl.close ()
		return scores
	
	def selectFeatures (self, name, selector, select, select_way, prune_df_min = 0, prune_df_max = 0, noprint = 0):
		self.si.log ("selecting [%s] using [%s:%s] for [%s]" % (select, selector, select_way, name))
		
		scores = self.scoringFeaures (selector, select_way, prune_df_min, prune_df_max)
		if not select:
			select = 1.0
		if select <= 1.0:
			select = int (len (scores) * select)
		else:
			select = int (select)
		if len (scores) < select:
			select = len (scores)
		
		if not noprint:
			c = 0
			for term, score in scores:
				c += 1
				tdf = self.writer.featureClassInfo (term, self.corpus)[0]				
				if c <= 200 or c >= (len (scores) - 200):
					self.si.log ("FEAT %5d. %s %2.3f (DF %d)" % (c, term, score, tdf))
				elif c == 201:
					self.si.log ("=========== Top 200, Bottom 200 ===========")
					
		self.si.log ("selecting within %d th features" % (select,))
		terms = scores [:select]
		terms = [x [0] for x in terms]
		
		fmap = dict ([(term, self.corpus.terms [term]) for term in terms])
		self.si.segments.features [name] = fmap	
		return len (terms)	
	
	def getFeatureSet (self, name):
		return self.si.segments.features	[name]	
	
	def listbydf (self, dfmin = 0, dfmax = 0):
		self.build (dfmin, dfmax, False)
	
	def closeIndexer (self):
		self.indexer.commit (1)
		self.indexer.merge ()
		self.indexer.close ()
		self.indexer = None			
		
		# reset parameters
		if self.si.hasSegmentsFile ():
			self.si.segments.parameters ["build"] = (-1, -1)
			self.si.segments.parameters ["train"] = {}
			if self.si.rebuild: 
				try: del self.si.segments.parameters ["optimized"]
				except KeyError: pass	
			self.si.flush (0)
					
	def build (self, dfmin = 0, dfmax = 0, commit = True, rebuild_index = False):
		if self.indexer:
			self.closeIndexer ()

		if not self._initialized:
			self.init ()
		
		if (dfmin, dfmax) == self.si.segments.parameters ["build"]:
			self.si.log ("same with previous parameters, need not rebuild")
			commit = False
		else:		
			self.si.log ("building corpus data")
			
		breader = self.writer.getSegmentBulkReader ()
		# get self.corpus pools and DF
		tindex = 0
		pruned = 0
		total = -1
		stat = {}
		while 1:
			try:
				ti = breader.advanceTermInfo ()
				if ti.fdno == self.writer.fdnoclass:
					self.corpus.pools.append (PoolInfo (ti.term, ti.df))
				elif ti.fdno == self.writer.fdnoterm:
					total += 1
					if dfmin and ti.df < dfmin:
						pruned += 1
					elif dfmax and ti.df > dfmax:
						pruned += 1	
					else:	
						self.corpus.terms [ti.term] = tindex
						stat [ti.term] = ti.df
						tindex += 1					
			except IndexError:
				break		
		
		temp = list(stat.items ())
		temp.sort (key = lambda x: x[1], reverse = True)		
		cc = 0
		for term, df in temp:
			cc += 1
			if cc <= 200 or cc >= (len (temp) - 200):
				self.si.log ("TERM %5d. DF %d %s" % (cc, df, term))
			elif cc == 201:
				self.si.log ("=========== Top 200, Bottom 200 ===========")	
		del stat
		del temp
		self.si.log ("%d (%d%%) terms pruned term dimension was shrinked: %d->%d" % (pruned, 1.*pruned/total*100, total, total-pruned))
		
		if rebuild_index:
			self.rebuildIndex (breader)						
			self.si.segments.parameters ["optimized"] = True
			self.si.flush (0)
		breader.close ()
		
		if not commit:
			self.writer.close ()		
			self.writer = None			
			return total, pruned
		
		self.writer.dirty = 1	
		self.corpus.pools.sort ()		
		# caching related with pools
		self.corpus.numpools = len (self.corpus.pools)
		pindex = 0
		poolno = {}
		for pool in self.corpus.pools:
			poolno [pool.name] = pindex
			pindex += 1
		self.corpus.poolno = poolno
		# caching related with pools--
		
		c = 0
		t = len (self.corpus.terms)
		for term in self.corpus.terms:
			if c % 1000 == 0:				
				self.si.log ("build corpus %d/%d" % (c, t))
			tdf, fi = self.writer.featureClassInfo (term, self.corpus)
			i = 0
			for df, tf in fi:
				self.corpus.pools [i].tf += tf
				i += 1
			c +=1
		
		self.writer.commit (self.corpus)		
		# sould be exactly here
		self.si.segments.parameters ["build"] = (dfmin, dfmax)		
		self.si.segments.parameters ["train"] = {}
		self.si.segments.features = {}
		self.writer.flush ()
		self.writer.close ()
		self.writer = None
		
		return total, pruned
		
	def untrain (cl_for = None):
		if cl_for in self.si.segments.features:
			del self.si.segments.features [cl_for]
			del self.si.segments.parameters ["train"][cl_for]
			self.si.flush ()	
		
	def train (self, cl_for = "default", **options):
		self.writer = composedsegmentwriter.ComposedSegmentFeaturesetWriter (self.si)
		if not self.corpus.numpools:
			self.corpus = self.writer.get_corpus ()
			
		selector = options.get ("selector", None)
		if not selector and cl_for == "default":
			raise ValueError("No default selector, be specify")
				
		if selector:
			orderby = options.get ("orderby", delune.SUM)
			select = float (options.get ("select", 1.0))
			prune_df_min, prune_df_max = options.get ("df_range", (0, 0))
			prune_df_min = options.get ("dfmin", 0)			
			prune_df_max = options.get ("dfmax", 0)
			noprint = options.get ("noprint", 0)
			
			if prune_df_min <= 1: prune_df_min = int (prune_df_min * len (self.corpus.terms))
			if prune_df_max <= 1: prune_df_max = int (prune_df_max * len (self.corpus.terms))
			
			numterms = self.selectFeatures (cl_for, selector, select, orderby, prune_df_min, prune_df_max, noprint)
		self.si.segments.parameters ["train"][cl_for] = options
		return numterms
				
	def optimize (self):
		if "optimized" in self.si.segments.parameters:
			raise AssertionError ("Already optimized")		
		if (-1, -1) == self.si.segments.parameters ["build"]:
			raise AssertionError ("Call build (dfmin, dfmax) first")
					
		dfmin, dfmax = self.si.segments.parameters ["build"]
		self.si.log ("optimizing corpus with DFMIN %d and DFMAX %d" % (dfmin, dfmax))
		self.build (dfmin, dfmax, True, True)		
		
		for cl_for, param in list (self.si.segments.parameters ["train"].items ()):
			self.si.log ("optimizing classifiers for %s" % (cl_for,))
			param ["noprint"] = 1
			self.train (cl_for, **param)
	
	
def usage ():
	print("""
Usage:
	%s [long option list...]
	--config=configuration file path
	--force
	--index
		sub option: --[bulk | append]	
	""" % sys.argv [0])
	sys.exit ()
	
	
def main (path, index = 0, rebuild = 0, force = 0):
	util.check_config (path)
	f = Trainer (path)
	if f.indexable (force):
		f.main (index, rebuild, force)


if __name__ == "__main__":
	import sys, getopt
	import profile

	try: argopt = getopt.getopt(sys.argv[1:], "", \
		["config=", "index", "bulk", "append", "force"])
	except:
		usage ()

	_config = None
	_rebuild = None
	_force = 0
	_index = 0

	for k, v in argopt[0]:
		if k == "--config":
			_config = v
		elif k == "--index":
			_index = 1
		elif k == "--bulk":
			_rebuild = 1
		elif k == "--append":
			_rebuild = 0
		elif k == "--force":
			_force = 1

	if not _config: usage ()
	if _index and _rebuild is None: usage ()

	main (_config, _index, _rebuild, _force)	



import sys
import os
from . import indexer, replicator
 
def help ():
    print ("usage: delune <command> [<options>]")
    print ("command:")
    print ("  index")
    print ("  replicate")    
    sys.exit ()
     
def main ():    
    try: 
        cmd = sys.argv [1]
    except IndexError:
        help ()
    sys.argv.pop (1)    
    if cmd == "index":
        indexer.main ()
    elif cmd == "replicate":
        replicator.main ()
    else:
        print ("unknown conmand")


if __name__ == "__main__":
    main ()



import delune
from delune import memory
import os
import shutil
import time
import pickle as pickle
from . import filesystem
from rs4 import logger as logs
from rs4.psutil import plock as _plock
from rs4.psutil import flock
from rs4.psutil import processutil
import socket
from functools import reduce
if os.name == "nt":
	import win32process

class SwapError (Exception):
	pass

class Options:
	pass
	
class Segments:
	def __init__ (self, alias = "", version = None):
		self.alias = alias
		self.version = version
		self.N = 0
		self.segmentInfo = {}


class CollectionSetup:
	exts = []
	segment_class = Segments
	
	def __init__ (self, indexdir, mode = 'r', analyzer = None, logger = None, plock = None, ident = None, version = 1):
		self.indexdir = indexdir
		self.mode = mode
		self.analyzer = analyzer
		self.logger = logger
		self.ident = ident
		self.fs = filesystem.FileSystem (self.indexdir)
		self.master = self.fs.getmaster ()		
		self.lock = flock.Lock (self.master)		
		self.version = version
	
		self.plock = plock # process lock at posix
		if self.plock is None and os.name == "posix":
			self.plock = _plock.Lock (os.path.join (self.master, ".plock"))
		
		self.rebuild = False
		if mode == "c":
			self.rebuild = True
			self.mode = "w"
		
		self.initial_newseg = 0
		self.options = Options ()

		if self.hasSegmentsFile ():
			self.read ()
			if self.rebuild:
				self.createSegments ()
		else:
			self.createSegments ()
		
		self._logger_created = False
		self._analyzer_created = False
		self._delune_init = False		
		
		if not self.logger:
			self.logger = delune.logger
			if not self.logger:
				self.logger = logs.screen_logger ()
				self.log ("logger is not provided, screen logger created", "warn")
				self._logger_created = True
		
		if self.mode == "r" and memory.memory is None: 
			self.log ("delune not inited, forcely init for single thread", "warn")
			delune.configure (1, self.logger)
			self._delune_init = True
		
		if self.analyzer is None:
			self.analyzer = delune.standard_analyzer (self.mode == delune.READ and 32 or 3000)			
			if self.mode == delune.READ:
				self.analyzer.setopt (ngram_no_space = 0)			
			self._analyzer_created = True
		
		if self.mode == delune.READ: # for multi-thread
			self.analyzer = delune.qualify_analyzer (self.analyzer)
	
	def set_ident (self, ident):
		self.ident = ident
		
	def close (self):
		self.lock.unlock ("using-%s" % (delune.PID or os.getpid ()))			
		if self._delune_init:
			self.log ("shutdown delune...", "info")
			delune.shutdown ()
			self._delune_init = False
						
		if self._analyzer_created:
			self.log ("close analyzer...", "info")
			self.analyzer.close ()
			self._analyzer_created = False
			
		if self._logger_created:			
			self.log ("close logger...", "info")
			self.logger.close ()
			self._logger_created = False		
	
	def gtVersion (self):
		return self.segments.version
	
	def getMajorVersion (self):
		return int (self.segments.version.split (".") [0])
		
	def notify (self):
		ll = list(self.segments.segmentInfo.keys ())
		ll.sort ()
		self.log ("segments updated... %s" % ll)	
					
	def write (self, fn = "segments"):
		self.segments.N = reduce (lambda x, y: x + y, list(self.segments.segmentInfo.values ()))				
		self.segments.alias = self.alias
		t = open (os.path.join (self.master, fn), "wb")
		pickle.dump (self.segments, t)
		os.fsync(t.fileno ())
		t.close ()
		self.log ("total indexed docuemts are %d" % self.segments.N)
			
	def read (self, fn = "segments"):	
		t = open (os.path.join (self.master, fn), "rb")
		self.segments = pickle.load (t)		
		t.close ()
		self.alias = self.segments.alias
		
		if type (self.segments.version) is str:
			self.version = 0
		
		if self.ident is None:
			self.ident = self.alias
		
		if self.mode == "r":
			self.setReadingSegments ()
		elif self.version == 0:
			raise SystemError ("version 0 had been deprecated, you can only read collection")
		self.initial_newseg = max (self.segments.segmentInfo.keys ()) + 1			
		
	def flush (self, new = 1):		
		if new:			
			segment = self.write ("segments.new")
		else:
			segment = self.write ()

		if new:
			self.log ("replace segments.new...")
			try:
				self.replaceFile ("segments")
			except:
				self.trace ()		
			self.notify ()
					
		if self.mode == "w":			
			self.removeDeletables ()

	def riaseSwapError (self, msg):
		raise SwapError("%s %s" % (self.alias, msg))

	def replaceFile (self, name):
		backup = os.path.join (self.master, name + ".bak")
		newone = os.path.join (self.master, name + ".new")
		replacement = os.path.join (self.master, name)

		try:
			try:
				os.remove (backup)
			except OSError as why:
				if why.errno != 2:
					self.riaseSwapError ("backup file cannot deleted")

			try:
				os.rename (replacement, backup)
			except OSError as why:
				if why.errno != 2:
					self.riaseSwapError ("backup failed")

			try:
				os.rename (newone, replacement)
			except:
				self.trace ()
				os.rename (backup, replacement)
				self.riaseSwapError ("replace failed. rollbacked")

			else:
				try: os.remove (backup)
				except: pass

		except:
			self.trace ()
			raise

		else:
			self.read ()
	
	def removeDeletables (self):
		if self.lock.isplocked ("duplicate"):
			return
		all = self.fs.segments ()
		used = self.getReadingSegmentsAllProcesses () + self.getSegmentList ()
		for seg in all:
			if seg not in used:
				self.removeSegmentFile (seg)

	def removeSegment (self, seg):
		try:
			del self.segments.segmentInfo [int (seg)]
		except KeyError:
			pass
	
	def removeSegmentFile (self, seg):
		try:
			self.fs.remove (seg)
		except:
			self.log ("segment %s remove failed" % seg, "fail")
			return 0

		self.log ("segment %s removed" % seg)		
		return 1
	
	def getSegmentsPath (self):
		return os.path.join (self.master, "segments")
		
	def getModfiedTime (self):
		try:			
			return os.stat (self.getSegmentsPath ()).st_mtime
		except (IOError, OSError): # no index yet
			return -1

	def createSegments (self):
		alias = os.path.split (self.master) [-1]
		self.segments = self.segment_class (alias, self.version)
		self.alias = self.segments.alias
		if self.ident is None:
			self.ident = self.alias

	def log (self, msg, type = "info"):
		self.logger (msg, type, self.ident)

	def trace (self):
		return self.logger.trace (self.ident)

	def getSegmentList (self):
		return list(self.segments.segmentInfo.keys ())

	def getLastSegment (self):
		segs = list(self.segments.segmentInfo.keys ())
		segs.sort ()
		return segs [-1]
		
	def getSegmentInfo (self):
		return self.segments.segmentInfo

	def getSegmentNumDoc (self, seg):
		try:
			return self.segments.segmentInfo [int (seg)]
		except KeyError:
			return 0

	def getNewSegment (self):
		if not self.segments.segmentInfo:
			return self.initial_newseg
		return max (self.segments.segmentInfo.keys ()) + 1

	def getSegmentInfo (self):
		return self.segments.segmentInfo

	def addSegment (self, seg, numDoc = 0):
		self.segments.segmentInfo [int (seg)] = numDoc

	def setReadingSegments (self):
		segs = self.getSegmentList ()
		if not segs: return
		segs = [str (x) for x in segs]
		self.lock.lock ("using-%s" % (delune.PID or os.getpid ()), ",".join (segs))		
	
	def getReadingSegments (self):
		segs = self.lock.lockread ("using-%s" % (delune.PID or os.getpid ()))
		if not segs: return []
		return [int (x) for x in segs.split (",")]
	
	def getReadingSegmentsAllProcesses (self):
		locks, errmsg = self.lock.locks ("using")
		d = {}
		for lname, ltime in locks:
			pid = int (lname.split ("-")[-1])
			isDeLuneServer = processutil.is_running (pid)
			if not isDeLuneServer:
				self.lock.unlock (lname)
				continue
					
			segs = self.lock.lockread (lname)
			if not segs: continue
			for seg in segs.split (","):
				d [int(seg)] = None
				
		return list(d.keys ())
		
	def isDeLuneServer (self, pid):
		proc = "/proc/%d/cmdline" % pid
		if os.path.isfile (proc):
			f = open (proc)
			exefilename = f.read ()
			f.close ()
			if exefilename.find ("python") != -1:
				return True
								
	def getN (self):
		return self.segments.N
		
	def getAlias (self):
		return self.alias
				
	def move (self, source, target, seg):
		self.copy (source, target, seg, True)
		
	def copy (self, source, target, seg, remove = False):
		for ext in self.exts:
			try:
				shutil.copyfile (os.path.join (source, str(seg) + "." + ext), os.path.join (target, str(seg) + "." + ext))
				if remove:
					os.remove (os.path.join (source, str(seg) + "." + ext))
			except IOError as why:
				if why [0] != 2:
					raise
	
	def clone (self, seg):
		current = self.fs.getsegmentpath (seg)
		newseg = self.getNewSegment ()
		newpath = self.fs.new (newseg)
		for ext in self.exts:
			src = os.path.join (current, "%d.%s" % (seg, ext))
			dest = os.path.join (newpath, "%d.%s" % (newseg, ext))
			if not os.path.isfile (src):
				continue
			if os.path.isfile (dest):
				os.remove (dest)
			shutil.copyfile (
				os.path.join (current, "%d.%s" % (seg, ext)),
				os.path.join (newpath, "%d.%s" % (newseg, ext))
			)
		return newpath, newseg
	
	def hasSegmentsFile (self):
		return os.path.isfile (os.path.join (self.master, "segments"))
	
	def removeSegmentsFile (self):
		os.remove (os.path.join (self.master, "segments"))
						
	def getopt (self, name = None, default = None, **karg):
		if name:
			try: return getattr (self.options, name)
			except AttributeError: return default
		
		for k, v in list(karg.items ()):
			try: return getattr (self.options, k)
			except AttributeError: return v
						
	def setopt (self, name = "", value = "", **karg):	
		if name:
			setattr (self.options, name, value)
		for k, v in list(karg.items ()):
			setattr (self.options, k, v)
		


import threading
import os
from rs4 import siesta
import delune
import time

class Commander:
	def __init__ (self, colopt, server, segments = 1, logger = None):
		self.colopt = colopt
		self.server = server
		self.segments = segments
		self.logger = logger
		self.api = siesta.API (self.server)
		
		self.apis = [None] * self.segments
		self.name = self.colopt ["name"]
		self.lk = threading.Lock ()
		self.T = 0		
	
	def get_api (self, segment):
		return self.apis [segment [0]]
	
	def get_info (self):
		return self.api.ops.list.get ().data
		
	def maybe_commit (self, api, message):
		with self.lk:
			self.T += 1
			if self.T % 100 == 0:			
				print (message, self.T)
				try:
					api (self.name).commit.get ()
				except:
					print ('commit error')
					
	def delete (self, segment, pk = None, query = None):
		api = self.get_api (segment)
		if query:
			r = api.ops (self.name).delete (query)
		elif pk:
			r = api.ops (self.name).documents (pk).delete ()
		else:
			raise ValueError ("Required pk or query")
			
		if r.data: raise SystemError (r.data)
		self.maybe_commit (api, pk and 'deleted')
	
	def post (self, segment, doc, pk = None):
		api = self.get_api (segment)
		while 1:
			r = None
			try:
				if pk:
					r = api.ops (self.name).documents (pk).patch (doc.as_dict ())
				else:
					r = api.ops (self.name).documents.post (doc.as_dict ())
			except:
				pass
			
			if not r or r.data:
				print ('error: %s, try after 1 sec...' % r)
				time.sleep (1)
			else:
				break
		
		self.maybe_commit (api, pk and 'updated' or 'inserted')
	
	def undo (self):
		if self.name in self.get_info () ["collections"]:	
			return
		
		r = self.api.ops (self.name).get (side_effect = 'undo')
		if r.data: raise SystemError (r.data)
		while 1:
			if self.name not in self.get_info () ["collections"]:	
				time.sleep (1)
			else:
				break

	def truncate (self):
		if self.name not in self.get_info () ["collections"]:	
			return
		r = self.api.ops (self.name).documents.delete (truncate_confirm=self.name)
		if r.data: raise SystemError (r.data)
	
	def commit (self):
		for api in self.apis:
			try:
				self.api.ops (self.name).commit.get ()
			except:
				pass			
	
	def register (self):
		if self.name not in self.get_info () ["collections"]:	
			r = self.api.ops (self.name).post (self.colopt)
			if r.data: raise SystemError (r)
		else:
			r = self.api.ops (self.name).patch (self.colopt)
			if r.data: raise SystemError (r.data)
		
		while 1:
			r = self.api.ops.list.get ()
			if self.name not in r.data ["collections"]:	
				time.sleep (1)
			else:
				break
				
	def unregister (self):
		if self.name not in self.get_info () ["collections"]:	
			return
			
		r = self.api.ops (self.name).delete (side_effect='data')
		if r.data: raise SystemError (r)
		while 1:
			if self.name in self.get_info () ["collections"]:	
				time.sleep (1)
			else:
				break

	def start (self, func, update = False):	
		ts = []		
		for i in range (1, self.segments):
			self.apis [i] = siesta.API (self.server)
			t = threading.Thread (target = func, args = (self, (i, self.segments), update))
			ts.append (t)
			t.start ()
		
		self.apis [0] = siesta.API (self.server)
		func (self, (0, self.segments), update)
		for t in ts:	
			t.join ()
		
		for api in self.apis:
			try:
				api.ops (self.name).commit.get ()
			except: 
				pass
		
			
def handle_commandline (colopt, server, segments = 1, index = None, delete = None):
	import sys
	
	commander = Commander (colopt, server, segments)	
	
	cmd = sys.argv [1]	
	if cmd == "unreg":
		commander.unregister ()	
	elif cmd == "reg":
		commander.register ()	
	elif cmd == "undo":	
		commander.undo ()	
	elif cmd == "truncate":	
		commander.truncate ()		
	elif cmd == "commit":	
		commander.commit ()	
	elif cmd == "index":	
		commander.start (index, False)	
	elif cmd == "update":	
		commander.start (index, True)		
	elif cmd == "delete":	
		commander.start (delete)	
	else:	
		print ('error: unknow command')



# 2017. 3. 13 by Hans Roh (hansroh@gmail.com)

from skitai.saddle import Saddle
import delune
from rs4 import pathtool
import os
import json
import codecs
import time
import shutil
import time

from decorative.admin import ui
from decorative.v1 import ops, pub

app = Saddle (__name__)
app.config.numthreads = 1
app.last_maintern = time.time ()

app.decorate_with (pub, mount = "/v1")
app.decorate_with (ops, mount = "/v1/ops", ns = "ops")
app.decorate_with (ui, mount = "/admin")

@app.route ("/status")
@app.permission_required (["index", "replica"])
def status (was):
	return was.status ()

@app.before_mount
def before_mount (wasc):
	app.config.numthreads = len (wasc.threads)
	permission_check_handler = wasc.app.config.get ("permission_check_handler")
	if permission_check_handler:
		app.permission_check_handler (permission_check_handler)
		
	delune.configure (app.config.numthreads, wasc.logger.get ("app"), 16384, 128)
	pathtool.mkdir (ops.getdir ("config"))
	for alias in os.listdir (ops.getdir ("config")):
		if alias.startswith ("-"):
			with codecs.open (ops.getdir ("config", alias), "r", "utf8") as f:
				colopt = json.loads (f.read ())
			for d in [ops.getdir ("collections", ops.normpath(d)) for d in colopt ['data_dir']]:
				if os.path.isdir (d):
					shutil.rmtree (d)
			os.remove (ops.getdir ("config", alias))
		elif alias.startswith ("#"):
			continue
		else:
			ops.load_data (alias, app.config.numthreads, wasc.plock)
	  
@app.umounted
def umounted (wasc):
	delune.shutdown ()

@app.before_request
def before_request (was):	
	with app.lock:
		last_maintern = was.getlu ("delune:collection") > app.last_maintern and app.last_maintern or 0
		if last_maintern:
			app.last_maintern = time.time ()
	
	if last_maintern:
		was.log ('collection changed, maintern...')
		for alias in os.listdir (ops.getdir ("config")):			
			if alias [0] in "#-":
				if delune.get (alias [1:]):
					delune.close (alias [1:])
			elif not delune.get (alias):
				ops.load_data (alias, app.config.numthreads, was.plock)			
			elif os.path.getmtime (ops.getdir ("config", alias)) > last_maintern:
				delune.close (alias)
				ops.load_data (alias, app.config.numthreads, was.plock)
	
	if was.request.args.get ('alias') and was.request.routed.__name__ != "collection":
		alias = was.request.args.get ('alias')
		if not delune.get (alias):
			return was.response.Fault ("404 Not Found", 40401, "resource %s not exist" % alias)



#!/usr/bin/env python

import os
import glob
from warnings import warn
import re

try:
	from setuptools import setup, Extension
except ImportError:	
	from distutils.core import setup, Extension
from distutils.sysconfig import get_python_lib		
import platform
import sys

with open('delune/__init__.py', 'r') as fd:
	version = re.search(r'^__version__\s*=\s*"(.*?)"',fd.read(), re.M).group(1)

if sys.argv[-1] == 'publish':
	buildopt = ['sdist', 'upload']
	if os.name == "nt":
		buildopt.insert (0, 'bdist_wheel')		
	os.system('python setup.py %s' % " ".join (buildopt))
	#os.system('twine upload dist/delune-%s*' % version)
	for each in os.listdir ("dist"):
		os.remove (os.path.join ('dist', each))
	sys.exit()

elif sys.argv[-1] == 'link':
	import site
	if os.name == "nt":
		linkdir = [each for each in site.getsitepackages() if each.endswith ("-packages")][0]		
	else:
		linkdir = [each for each in site.getsitepackages() if each.find ("/local/") !=- 1 and each.endswith ("-packages")][0]		
	target = os.path.join (os.path.join (os.getcwd (), os.path.dirname (__file__)), "delune")
	link = os.path.join (linkdir, "delune")
	if os.name == "nt":
		os.system ("mklink /d {} {}".format (link, target))
	else:
		os.system ("ln -s {} {}".format (target, link))	
	sys.exit ()
	
modules = []
osbit, _dummy = platform.architecture()
dira = osbit == "32bit" and "x86" or "x64"

if os.name == "nt":
	include_dirs = [
		"delune/win32inclib/zlib/%s" % dira, 
		"delune/win32inclib/pthread2/%s" % dira
	]
	library_dirs = include_dirs
	libraries = ["zlib", "pthreadVC2"]
		
else:
	include_dirs = ["/usr/include/x86_64-linux-gnu", "/usr/include"]
	library_dirs = ["/usr/lib/x86_64-linux-gnu", "/usr/lib"]
	libraries = ["z", "pthread"]

module = Extension(
	'delune._delune',
	sources = [
		'delune/extension/delune/core.c', 
		'delune/extension/delune/analyzer/analyzer.c',		
		'delune/extension/delune/analyzer/stopword.c', 
		'delune/extension/delune/analyzer/endword.c', 
		'delune/extension/delune/analyzer/formalizer.c', 		
		'delune/extension/delune/analyzer/stem.c',
		'delune/extension/delune/analyzer/stem_de.c',
		'delune/extension/delune/analyzer/stem_fr.c',
		'delune/extension/delune/analyzer/stem_it.c',
		'delune/extension/delune/analyzer/stem_fi.c',
		'delune/extension/delune/analyzer/stem_es.c',
		'delune/extension/delune/analyzer/stem_hu.c',
		'delune/extension/delune/analyzer/stem_pt.c',
		'delune/extension/delune/analyzer/stem_sv.c',
		'delune/extension/delune/analyzer/stem_ar.c',
		'delune/extension/delune/analyzer/removeaccents.c',
		'delune/extension/delune/index/sort.c', 
		'delune/extension/delune/index/osutil.c', 
		'delune/extension/delune/index/search.c', 
		'delune/extension/delune/index/compfunc.c',
		'delune/extension/delune/index/heapsort.c', 
		'delune/extension/delune/index/termhashtable.c', 
		'delune/extension/delune/index/generichash.c',
		'delune/extension/delune/index/bfile.c', 
		'delune/extension/delune/index/zip.c',
		'delune/extension/delune/index/mempool.c',
		'delune/extension/delune/index/ibucket.c',
		'delune/extension/delune/index/util.c',
		'delune/extension/delune/mod_analyzer.c',		
		'delune/extension/delune/mod_util.c', 
		'delune/extension/delune/mod_int.c', 
		'delune/extension/delune/mod_bits.c', 
		'delune/extension/delune/mod_mergeinfo.c', 
		'delune/extension/delune/mod_document.c',
		'delune/extension/delune/mod_posting.c', 
		'delune/extension/delune/mod_termtable.c',
		'delune/extension/delune/mod_terminfo.c',
		'delune/extension/delune/mod_sortmap.c',   
		'delune/extension/delune/mod_memorypool.c',		
		'delune/extension/delune/mod_selector.c',		
		'delune/extension/delune/mod_classifier.c',
		'delune/extension/delune/mod_dbint.c',
		'delune/extension/delune/mod_calculator.c',
		'delune/extension/delune/mod_binfile.c',		
		'delune/extension/delune/mod_sgmlparser.c',
		'delune/extension/delune/mod_compute.c',
	],
	include_dirs = include_dirs,
	library_dirs = library_dirs,
	libraries = libraries
)
modules.append (module)

packages = [
	'delune',
	'delune.analyzers',
	'delune.analyzers.util',
	'delune.searcher',	
	'delune.searcher.segment',
	'delune.classifier',	
	'delune.classifier.classifiers',
	'delune.classifier.segment',	
	'delune.cluster',
	'delune.util',	
	'delune.bin',
	'delune.cli',
	'delune.export',
	'delune.export.skitai',
	'delune.export.skitai.decorative',	
	'delune.export.skitai.decorative.admin',
	'delune.export.skitai.decorative.v1',
]

package_dir = {
	'delune': 'delune'
}
   
data_files = [
	"extension/delune/*.*",
	"extension/delune/analyzer/*.*",
	"extension/delune/index/*.*",	
	"extension/delune/win32/*.*",	
	"win32inclib/pthread2/x64/*.*",
	"win32inclib/pthread2/x86/*.*",	
	"win32inclib/zlib/x64/*.*",
	"win32inclib/zlib/x86/*.*",
	"export/skitai/templates/*.*",	
]

if os.name == "nt" and ("build" in sys.argv or "install" in sys.argv):
	import shutil		
	try:
		if os.path.isfile ("delune/pthreadVC2.dll"):
			os.remove ("delune/pthreadVC2.dll")
		shutil.copy ("delune/win32inclib/pthread2/%s/pthreadVC2.dll" % dira, "delune/pthreadVC2.dll")		
		pass
	except OSError:
		pass	
	data_files.extend (["pthreadVC2.dll"])

package_data = {
	"delune": data_files
}

classifiers = [
  'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
  'Development Status :: 3 - Alpha',
  'Environment :: Console',	
	'Topic :: Software Development :: Libraries :: Python Modules',
	'Intended Audience :: Developers',
	'Intended Audience :: Science/Research',
	'Programming Language :: Python',
	'Programming Language :: Python :: 3',
	'Topic :: Text Processing :: Indexing'	
]


install_requires = ["skitai"]

with open ('README.rst', encoding='utf-8') as f:
	long_description = f.read()
	
setup (
	name = 'delune',
	version = version,
	author = 'Hans Roh',
	description='DeLune Python Object Storage and Search Engine',
	long_description = long_description,
	author_email = 'hansroh@gmail.com',
	url = 'https://gitlab.com/hansroh/delune',
	packages=packages,
	package_dir=package_dir,
	package_data = package_data,
	entry_points = {
        'console_scripts': [
			'delune=delune.bin.delune:main',			
		],
    },
	license='GPLv3',
	platforms = ["posix", "nt"],
	download_url = "https://pypi.python.org/pypi/delune",
	install_requires = install_requires,
	classifiers=classifiers,
	ext_modules = modules,
	zip_safe = False
)




"""
Install Mecab and Dictionary
---------------------------------
apt install curl autoconf g++ python3-dev
$ bash < (curl -s https://raw.githubusercontent.com/konlpy/konlpy/master/scripts/mecab.sh)

https://bitbucket.org/eunjeon/mecab-ko-dic/downloads/
wget https://bitbucket.org/eunjeon/mecab-ko-dic/downloads/mecab-ko-dic-2.0.3-20170922.tar.gz

tar -zxfv mecab-ko-dic-XX.tar.gz
cd mecab-ko-dic-XX
./configure
make
sudo make install

Install JDK for Java morphers
---------------------------------
apt-get install default-jdk

Install konlpy
----------------
pip3 install JPype1-py3 konlpy

(Optional) Change System Locale
---------------------------------
apt install language-pack-ko
sudo locale-gen ko_KR.UTF-8

sudo vi /etc/default/locale   
 - LANG=ko_KR.UTF-8
source /etc/default/locale
"""

from . import standardAnalyzer
import re

class Analyzer (standardAnalyzer.Analyzer):
	defaults = {	
		"strip_html": 0,
		"make_lower_case": 0,
		"stopwords": [],
		"contains_alpha_only": 0,
		"dicpath": '/usr/local/lib/mecab/dic/mecab-ko-dic'
	}
	
	def __init__ (self, max_term = 8, numthread = 1, **karg):
		self.stopwords = {}
		standardAnalyzer.Analyzer.__init__ (self, max_term, numthread, **karg)
		
	def createTSAnalyzers (self, num):
		from konlpy.tag import Mecab
		
		#I don't know Mecab is thread-safe
		for i in range (num):
			self.ats.append (Mecab (self.options.get ("dicpath")))
	
	#------------------------------------------------
	# exported methods
	#------------------------------------------------				
	def index	(self, document, lang = "en"):
		document = self.preprocess (document)
		d = {}
		t = 0
		for pos, term in enumerate (self.ats [self.get_tid ()].morphs (document)):
			if term in self.stopwords:
				continue
			if term not in d:
				d [term] = []
			d [term].append (pos)	
			t += 1
			if t > self.max_term:
				break			
		return d
	
	def tokenize (self, document, lang = "en"):
		if self.options.get ("make_lower_case"):
			document = document.lower ()			
		return [term for term in self.ats [self.get_tid ()].morphs (document) if term not in self.stopwords]
	query = tokenize
	
	def nouns (self, document):
		return self.ats [self.get_tid ()].nouns (document) [:self.max_term]

	def pos (self, document):
		return self.ats [self.get_tid ()].pos (document) [:self.max_term]
		
	def count_stopwords (self, document, lang = "en"):
		document = self.preprocess (document)
		terms = self.tokenize (document)
		c = 0
		for term in terms:
			if term in self.stopwords:
				c += 1
		return c, len (terms)
		
	#------------------------------------------------
	# handle options 
	#------------------------------------------------		
	def set_stopwords (self, wordlist):
		if not wordlist: return
		self.stopwords = dict ([(x, None) for x in wordlist])
			
	def set_endwords (self, wordlist):
		raise AttributeError
	
	def set_stem_level (self, stem_level):
		raise AttributeError
		
	def set_ngram_no_space (self, flag):
		raise AttributeError
	
	def set_stopwords_case_sensitive (self, flag):
		raise AttributeError
				
	def set_ngram (self, ngram):
		raise AttributeError
	
	def set_biword (self, flag):
		raise AttributeError
		


import re
import urllib.request, urllib.parse, urllib.error

class QuerySet:
	rx_col = re.compile ("\[info:(.+?)\]")
	def __init__ (self, log, loop = 1, max = 0):
		self.loop = loop
		self.f = open (log)
		self.max = max
		self.c = 0
	
	def close (self):
		self.f.close ()
	
	def get_uri (self):
		args = self.get ()
		return "/rpc2/bladese?col=%s&qs=%s&offset=%s&fetch=%s&sort=%s&summary=%s&analyze=%s&delete=%s&field%s" % args
		
	def get (self):
		while 1:
			line = self.f.readline ()
			if not line:
				if self.loop:
					self.f.seek (0)
					self.f.readline ()
				else:
					return None	
			
			s = line.find ("QUERY ")
			if s == -1: continue
			
			match = self.rx_col.search (line)
			if match:
				col = match.group (1)
			else: 
				continue
			
			line = line [s + 6:].strip ()
			k = line.split ("/")		
			qs = urllib.parse.quote ("/".join (k [:-7]))
			args = (col, qs) + tuple (k [-7:])
			if len (args) != 9: continue
			
			self.c += 1
			if self.max and self.c > self.max: return None	
				
			return args


class QueryResultSet:
	rx_col = re.compile ("\[info:(.+?)\]")
	def __init__ (self, log):
		self.f = open (log)
		self.rset = {}
	
	def close (self):
		self.f.close ()
			
	def build (self):
		while 1:
			line = self.f.readline ()
			if not line:
				return None	
			
			s = line.find ("RESULT ")
			if s == -1: continue
			
			match = self.rx_col.search (line)
			if match:
				col = match.group (1)
			else: 
				continue
			
			line = line [s + 6:].strip ()			
			q, r = line.split ("=>", 1)
			r = r.split ("/", 4)			
			self.rset [col + "/" + q] = int (r [2])
			



import delune
from delune import binfile
import queue
from delune.searcher.indexer import ExitNow
from rs4 import pathtool
import os, glob, json
import getopt, sys, time
from rs4.psutil import service, daemon_class
import threading
import shutil

class	Indexer (daemon_class.DaemonClass):
	NAME = "indexer"

	def __init__ (self, logpath, varpath, consol, config):
		self.config = config
		self.lock = threading.RLock ()
		self.enter_shutdown = False
		self.q = queue.Queue ()
		self.working = {}
		self.ts = []
		daemon_class.DaemonClass.__init__ (self, logpath, varpath, consol)
	
	def add_document (self, indexer, doc):
		if self.enter_shutdown:
			raise daemon_class.ExitNow
				
		document = delune.document ()	
		document.documents (doc ['documents'])
		if 'snippet' in doc and doc ['snippet']:		
			document.snippet (*tuple (doc ['snippet']))
	
		for name, opt in doc ['fields'].items ():
			if name == "_id":			
				data = str (opt)
				lang = 'un'
				ftype = delune.STRING
				option = []
				
			else:	
				data = opt.get ('data')
				if type (data) is str and data.startswith ("$document"):
					tree = data.split (".")
					if tree [0][9:] == "":
						index = 0
					else:	
						try: index = int (tree [0][9:])
						except ValueError: 
							raise TypeError ("document index is wrong")
							
					val = doc ['documents'][index]
					for dep in tree [1:]:
						if type (val) is dict:
							val = val.get (dep)
						else:
							try: val = val [int (dep)]
							except IndexError: val = None															
						if not val:
							break
					data = val
										
				ftype = opt.get ('type')
				if not ftype:
					raise TypeError ('Type of field is required')
				option = opt.get ('option', [])				
				lang = opt.get ('lang', 'un')
				
			document.field (name, data, ftype, lang, option = option)
		indexer.add_document (document)
	
	def getdir (self, *d):
		return os.path.join (self.config ["resource_dir"], "delune", *d)
	
	def delete (self, alias, col, opt, deletables):
		opt ['searcher']['remove_segment'] = 0
		searcher = col.get_searcher (**opt ['searcher'])
		if searcher.si.getN () == 0:
			searcher.close ()
			return 0
		
		deleted = 0
		for qs in deletables:
			qs ["commit"] = False
			deleted += searcher.delete (**qs)["total"]
			
		searcher.commit ()	
		searcher.close ()	
		self.log ('deleted %d documents' % deleted, "info", alias)	
		return deleted
			
	def index_collection (self, alias, colopt, col):
		queue_dir = colopt ['data_dir']
		if type (queue_dir) is list:
			queue_dir = queue_dir [0]
		
		queue_dir = os.path.join (queue_dir, ".que")
		queues = [(os.path.basename (q), os.path.getmtime (q)) for q in glob.glob (os.path.join (queue_dir, 'que.*')) if not q.endswith (".lock")]
		
		if not queues:
			return
				
		self.log ("indexing start", "echo", alias)
		# reverse sort for trucate
		queues.sort (key = lambda x: x [-1], reverse = True)
		index = 0
		for gfile, mtime in queues:
			if gfile == "que.truncate":
				break
			index += 1
		
		if index < len (queues):
			self.log ("found trcuate request", "info", alias)
			for qfile, mtime in queues [index:]:
				os.remove (os.path.join (queue_dir, qfile))
			queues = queues [:index]
			indexer = col.get_indexer (**colopt ['indexer'])
			indexer.truncate ()	
			indexer.close ()
		
		deleted = 0
		deletables = []
		# resort from old one
		queues.reverse ()	
		for qfile, mtime in queues:
			bf = binfile.BinFile (os.path.join (queue_dir, qfile), "r")
			while 1:
				try: 
					cmd = bf.readVInt ()
				except OSError:
					bf.close ()
					break
				try:
					doc = json.loads (bf.readZBytes ().decode ("utf8"))
				except (OSError, MemoryError):
					bf.close ()
					break
				
				if cmd == 1:
					qs = doc.get ("query")
					deletables.append (qs)
					if len (deletables)	== 10000:
						# too many, commit forcely
						deleted += self.delete (alias, col, colopt, deletables)
						deletables = []
		
		if deletables:
			deleted += self.delete (alias, col, colopt, deletables)
		if deleted:	
			self.log ('deleted total %d documents' % deleted, "info", alias)
			colopt ['indexer']["optimize"] = 1
			colopt ['indexer']["force_merge"] = 1			
		indexer = col.get_indexer (**colopt ['indexer'])	
		# mannual commit mode, ignore memory usage
		indexer.set_autocommit (False)
		
		total_indexed_docs = 0
		indexed_docs = 0
		indexed_ques = []
		for qfile, mtime in queues:
			bf = binfile.BinFile (os.path.join (queue_dir, qfile), "r")
			ndoc = 0
			while 1:
				try: cmd = bf.readVInt ()				
				except OSError:
					bf.close ()
					break
				
				if cmd == 1:	 # just delete
					bf.readZBytes ()
					continue
				
				try:
					doc = json.loads (bf.readZBytes ().decode ("utf8"))				
				except (OSError, MemoryError):
					bf.close ()
					break
				
				self.add_document (indexer, doc)
				indexed_docs += 1
				ndoc += 1
				
			self.log ('added %d documents from %s' % (ndoc, qfile), "info", alias)
			indexed_ques.append (os.path.join (queue_dir, qfile))
			if self.maybe_commit (alias, indexer, indexed_ques, indexed_docs):
				total_indexed_docs += indexed_docs
				indexed_docs = 0
				indexed_ques = []
	
		self.maybe_commit (alias, indexer, indexed_ques, indexed_docs, force = True)
		if indexed_docs:
			total_indexed_docs += indexed_docs 
			self.log ('added total %d documents added' % total_indexed_docs, "info", alias)	
		indexer.close (optimize = deleted > 100)
	
	def maybe_commit (self, alias, indexer, indexed_ques, indexed_docs, force = False):
		if force or indexer.is_memory_over ():
			# commit and delete queue for preventing duplication		
			if indexed_docs:
				indexer.commit ()
			for qpath in indexed_ques:
				self.log ('remove indexed que %s' % os.path.basename (qpath), "info", alias)
				os.remove (qpath)
			return 1
		return 0
	
	def index (self):
		while 1:
			alias = self.q.get ()
			if not alias: return
			if self.enter_shutdown: return	
			with self.lock: self.working [alias] = None
			self.log ("checking...", "echo", alias)
			col = None
			try:
				path = self.getdir ("config", alias)
				if alias [0] in "#-":
					continue
				with open (path, "r") as f:
					colopt = json.loads (f.read ())				
				if colopt.get ("version", 1) == 0 or "ro" in colopt.get ("permission", []):
					self.log ("read only collection, skip", "info", alias)
					continue
						
				colopt ['data_dir'] = [self.getdir ("collections", os.path.normpath(d)) for d in colopt ['data_dir']]
				
				col = delune.collection (
				  indexdir = colopt ['data_dir'],
				  mode = delune.APPEND,
				  analyzer = delune.standard_analyzer (colopt ['analyzer'].get ("max_terms", 3000), 1, **colopt ['analyzer']),
				  version = colopt.get ("version", 1)
				)
				if col.lock.isplocked ("replicate"):
					self.log ("on replicating, maybe next", "warn", alias)
					continue
				self.index_collection (alias, colopt, col)
				col.removeDeletables ()
				
			except Exception as why:				
				if not isinstance (why, daemon_class.ExitNow):
					self.trace ()
			
			with self.lock: del self.working [alias]
			if self.enter_shutdown: return
	
	def hterm (self, signum, frame):
		self.log ("got SIGTERM, entering shutdown process...", "echo")		 	
		self.enter_shutdown = True
		for i in range (self.config.get ("threads", 1) * 3):
			self.q.put (None)
					
	def setup (self):
		self.bind_signal (self.hterm)
		delune.configure (self.config.get ("threads", 1), self.logger)
		
	def wait_for_interval (self):
		interval = self.config.get ("interval", 0)		
		if interval == 0:
			return False
		for i in range (interval):
			if self.enter_shutdown:
				return False
			time.sleep (1)
		return True
	
	def maintern (self, config_dir):
		collections_dir = os.path.join (os.path.dirname (config_dir), "collections")
		for alias in os.listdir (config_dir):
			if alias [0] not in "-":
				continue
			self.log ("removing local collection", 'warn', alias)
			config_path = self.getdir ("config", alias)
			with open (config_path) as f:
				colopt = json.loads (f.read ())
			for each in [os.path.join (collections_dir, os.path.normpath(d)) for d in colopt ['data_dir']]:
				if os.path.isdir (each):
					shutil.rmtree (each)
			os.remove (config_path)	
			self.log ("local collection has been removed", 'warn', alias)
					
	def threading (self, func):
		self.ts = []
		for i in range (self.config.get ("threads", 1)):	
			t = threading.Thread (target = func)
			self.ts.append (t)
			t.start ()
		
	def reque (self, aliases):		
		if not self.q.empty ():
			return
		for alias in aliases:
			if [ex for ex in self.config.get ("exclude", []) if alias.startswith (ex)]:
				continue
			with self.lock:
				if alias in self.working:
					continue
			self.q.put (alias)
		
		if self.config.get ("interval", 0) == 0:
			for t in self.ts:
				self.q.put (None)
				t.join ()
	
	def run (self):
		pathtool.mkdir (self.getdir ("config"))
		if self.config.get ("threads", 1) == 0:
			self.reque (os.listdir (self.getdir ("config")))
			self.q.put (None)
			return self.index ()
		
		self.threading (self.index)
		while 1:	
			self.maintern (self.getdir ("config")) 
			self.reque (os.listdir (self.getdir ("config")))		
			if not self.wait_for_interval ():
				break
				

def usage ():
	print ("""usage: 
  delune index [options] [start|stop|status|restart]

options:
  -d: daemonize
  -e, --exclude: <alias,...>
  -v, --resource-dir: <resource-dir>
  -i, --interval: <int> interval seconds 
  -t, --threads: <int> 
      --help 

example:
  delune index -i 300 -vs /home/ubuntu/deluned/resource
	""")
	sys.exit ()

def remove_end_slash (url):
	while url:
		if url [-1] != "/":
			return url
		url = url [:-1]	
		
def parse_argv (sopts, lopts, usage_): 
	argopt = getopt.getopt(sys.argv[1:], sopts, lopts)
	
	config = {}
	try: cmd = argopt [1][0]
	except: cmd = None
	for k, v in argopt [0]:
		if k == "-o" or k == "--origin":
			config ["origin"] = remove_end_slash (v)
		elif k == "-v" or k == "--resource-dir":
			config ["resource_dir"] = v
		elif k == "--help":
			usage_ ()	
		elif k == "-t" or k == "--threads":
			config ["threads"] = int (v)
		elif k == "-e" or k == "--exclude":
			config ["exclude"] = [each.strip () for each in v.split (",")]	
		elif k == "-i" or k == "--interval":
			config ["interval"] = int (v)
		elif k == "-d":
			cmd = "start"
	return cmd, config, not cmd	

VARPATH = "/var/tmp/delune/indexer"
LOGPATH = "/var/log/delune/indexer"

def index (model_root):
	 f = Indexer (LOGPATH, VARPATH, True, {"threads": 0, "resource_dir": model_root})
	 f.start ()
	 f.close ()

def main ():
	cmd, config, console = parse_argv ("dv:o:t:i:e:", ["exclude=", "origin=", "threads=", "interval=", "resource-dir", "help"], usage)
	if not cmd or cmd in ("start", "restart"):
		assert "resource_dir" in config, "resource root required"
	assert "origin" not in config, "invalid option: origin"
		
	pathtool.mkdir (VARPATH)
	servicer = service.Service ("delune/{}".format (Indexer.NAME), VARPATH)
	if cmd and not servicer.execute (cmd):
		return
	if console and servicer.status (False):
		raise SystemError ("daemon is running")
	
	s = daemon_class.make_service (Indexer, LOGPATH, VARPATH, console, config)
	s.start ()
	s.close ()
	
if __name__ == "__main__":
	main ()




from rs4 import siesta
from . import local
from rs4 import logger
import time

class Documents (local.Documents):
    def __init__ (self, name, addr, logger):
        self.addr = addr
        self.logger = logger
        self.api = siesta.API (self.addr)
        self.name = name
        self.counts = [0, 0]
    
    def __enter__ (self):
        return self
    
    def __exit__ (self, type, value, tb):
        pass
            
    def add (self, doc):
        if doc._id:
            return self.update (doc)    
        self.logger ("#{} add document".format (self.counts [0]), "info", self.name)
        self.counts [0] += 1
        self.api.ops (self.name).documents.post (doc.as_dict ())
    
    def update (self, doc):
        assert doc._id, "_id required"
        self.logger ("#{} upsert document ({})".format (self.counts [0], doc._id), "info", self.name)
        self.counts [0] += 1
        return self.api.ops (self.name).documents (doc._id).patch (doc.as_dict ())
    
    def delete (self, id):
        self.logger ("#{} delete document ({})".format (self.counts [1], id), "info", self.name)
        self.counts [1] += 1
        return self.api.ops (self.name).documents (id).delete ()
    
    def qdelete (self, qs):
        self.api.ops (self.name).documents.delete (q = qs)
    
    def truncate (self, name):
        self.api.ops (self.name).documents.delete (truncate_confirm = name)
        
    def search (self, qs):
        return self.api (self.name).search.get (q = qs)
    
    def guess (self, qs):
        return self.api (self.name).guess.get (q = qs)
    
    def commit (self):
        self.api.ops (self.name).commit.get ()
    
    def index (self):
        raise AttributeError 
    
    def rollback (self):
        self.api.ops (self.name).rollback.get ()
        

class Collection:
    def __init__ (self, name, config, addr, logger = None):
        self.addr = addr
        self.logger = logger
        self.api = siesta.API (self.addr)
        self.name = name
        self.config = config
        self.documents = Documents (self.name, self.addr, logger)
    
    def _wait (self, flag = True):
        for i in range (60):
            if self.is_active () is flag:
                break
            else:
                time.sleep (1)
            
    def is_active (self):
        return self.name in self.api.ops.list.get ().data ["collections"]
    
    def save (self):
        if self.is_active ():
            self.api.ops (self.name).patch (self.config)            
        else:
             self.api.ops (self.name).post (self.config)   
        self._wait (True)
            
    def drop (self, include_date = False):
        if not self.is_active ():
            return
        if include_date:
            self.api.ops (self.name).delete (side_effect = 'data')
        else:
            self.api.ops (self.name).delete ()    
        self._wait (False)
    
    def undrop (self):
        if self.is_active ():
            return
        self.api.ops (self.name).get (side_effect = 'undo')
        self._wait (True)
    
    
class Delune:
    def __init__ (self, addr):
        self.addr = addr
        self.logger = logger.screen_logger ()
        self.api = siesta.API (self.addr)
        self.cols = self.lscol ()
         
    def lscol (self):
        return self.api.ops.list.get ().data ["collections"]
    
    def load (self, name):
        if name not in self.cols:
            raise NameError ("collection not found")
        config = self.api.ops (name).config.get ().data            
        return Collection (name, config, self.addr, self.version)
    
    def create (self, name, data_dirs = [], version = 1, **kargs):
        config = local.make_config (name, data_dirs, version, **kargs)
        return Collection (name, config, self.addr, self.logger)
    